/* Inomads Apache module
 *
 * (C) copyright 2009, Business Basic Systems Corporation
 * Licensed to PVX Plus Technologies Ltd., Ontario, Canada
 * All rights reserved worldwide.
 * Unauthorized use of this code is strictly prohibited.
 *
 * This is the Apache load module used by iNomads
 */

#include <httpd.h>
#include <http_protocol.h>
#include <http_config.h>
#include <http_core.h>
#include <apr_strings.h>
#include <string.h>
#include <http_log.h>


#if defined(_WINDOWS)
#define NL				"\r\n"
#define PXPLUS_EXE		"pxplus.exe"
#else
#define NL		"\n"
#define PXPLUS_EXE		"pxplus"
#include <unistd.h>

int MyMemicmp(const void *s1, const void *s2, size_t n)
{	int	nOfst;
    int nDiff;

    unsigned char *p1 = (unsigned char *)s1;
    unsigned char *p2 = (unsigned char *)s2;

    for (nOfst = 0; nOfst < n; nOfst++ )
	{	nDiff = toupper(p1[nOfst]) - toupper(p2[nOfst]);
		if (nDiff != 0)return(nDiff);
    }
	return 0;
}

#define memicmp MyMemicmp

#endif

#define MAX_PATH_NAME		256

static int UnEscapeURL(char *pVal);

struct CONFIG
{	apr_table_t			*tblLinks;		/* Quick look up table for session/link info */
	char				*pDir;			/* Application directory */
	char				*pExe;			/* PxPlus exec */
	char				*pLib;			/* PxPlus library */
	char				*pHtml;			/* PxPlus syshtml directory */
	char				*pLaunchIP;		/* Launcher IP address */
	int					nLaunchPort;	/* Launcher Port number */
	int					nTimeout;		/* Timeout in seconds */
	char				*pLaunchPswd;	/* Launcher Password */
	char				*pRmtRoot;		/* LaunchPort Remote webroot */
};

struct INOMADS
{	request_rec			*r;
	apr_table_t			*tblForm;
	apr_table_t			*tblArgs;
	apr_pool_t			*sysPool;		/* Server process pool */
	struct CONFIG		*pConf;			/* Config info (in SysPool) */
	char				*pContent;		/* Submitted contents */
	char				*pType;			/* Content type */
	char				*pDocRoot;		/* Document root -- as we see it */
	char				*pRmtRoot;		/* Document root -- as a launch port would get it */
	char				*pSession;		/* Session ID */
	char				*pLink;			/* Link pathname (TCP addr;port presently) */
	char				*pSesDiry;		/* Session temp directory */
	char				*pInid;			/* Workstation identifier */		
	char				*pAjax;			/* Ajax wrapup command */
	char				*pMeta;			/* Meta data (not used) */
	char				*pMsg;			/* Message text */
	char				*pFile;			/* Pointer to uploaded file (only one allowed) */
	char				*pFileName;		/* Users workstation file name */
	long				lContent;		/* Length of contents */
	int					bUpload;		/* Content is an Upload */
	int					rcSts;			/* Error code return value */
};

struct TBLINFO
{	int					nKeys;		/* Number of keys */
	int					lenKeys;	/* Total length of all the keys */
	int					lenData;	/* Total length of all the data */
	char				*pKeys;		/* Key name list */
	char				*pData;		/* Data list */
};

#define Request		pNomads->r
#define TmpPool		pNomads->r->pool
#define SysPool		pNomads->sysPool
#define RetCode		pNomads->rcSts

static int Load_Config(struct INOMADS *pNomads);
static int Parse_Args(struct INOMADS *pNomads); 
static int Read_Contents(struct INOMADS *pNomads);
static int Parse_Contents(struct INOMADS *pNomads);
static int Parse_Upload(struct INOMADS *pNomads);

static int Load_session(struct INOMADS *pNomads);
static int Spawn_Direct(struct INOMADS *pNomads);
static int Spawn_ViaTcp(struct INOMADS *pNomads);
static int Load_Inid(struct INOMADS *pNomads);
static void Exchg_data(struct INOMADS *pNomads);

static int Send_Error(struct INOMADS *pNomads, char *pMsg);
static void Send_SysHtml(struct INOMADS *pNomads, char *sHtmlFile);

static apr_table_t *TblParse(struct INOMADS *pNomads, const char *data);
static int CountArg(void* pCnt, const char* pKey, const char* pValue);
static int PrintArg(void* pFile, const char* pKey, const char* pValue);

static int inomads_handler(request_rec* r)
{	struct INOMADS		iNomads;
	char				*pDlm;
	
	if (!r->handler) return DECLINED;
	
	pDlm = strrchr(r->uri, '/');
	if (pDlm == NULL) 
		pDlm = (char *)r->uri;
	else
		pDlm++;

	if (apr_strnatcasecmp(pDlm, "inomads"))
	{	if (apr_strnatcasecmp(pDlm, "bmobile"))
			return DECLINED;
	}

	if ( r->method_number != M_GET )
	{	if ( r->method_number != M_POST )return HTTP_METHOD_NOT_ALLOWED ;  /* Reject other methods */
	}
	
	r->no_cache = TRUE;
	
	ap_set_content_type(r, "text/html;charset=ascii") ;
	memset(&iNomads, 0, sizeof(iNomads));

	iNomads.r = r;
	iNomads.sysPool = r->server->process->pconf;		/* Config pool -- cleared on restart */

	iNomads.pDocRoot = apr_pstrdup(r->pool, r->filename);
	pDlm = strrchr(iNomads.pDocRoot, '/');
	if (pDlm) *pDlm = 0;

	if (!Load_Config(&iNomads)) return OK;

	if (iNomads.pConf->pRmtRoot[0] != 0)
		iNomads.pRmtRoot = iNomads.pConf->pRmtRoot;
	else
		iNomads.pRmtRoot = iNomads.pDocRoot;
	
	Parse_Args(&iNomads);

	if (Read_Contents(&iNomads) &&
		Parse_Contents(&iNomads) &&
		Load_session(&iNomads))
			Exchg_data(&iNomads);
	
	return OK;
}

/* Load_Config(pNomads)
 *
 *		Recall the config information from the system pool or create it first time if
 *		not present.
 *		This routine saves the config into the system memory so that subsequent requests
 *		will not require and file IO.
 */
static int Load_Config(struct INOMADS *pNomads)
{	apr_file_t		*pfConf;
	char			szLine[MAX_PATH_NAME+100];
	char			szPath[MAX_PATH_NAME];
	char			*pDlm;
	char			*pTemp;
	
	strcpy(szLine, "[Inomads]");
	strcat(szLine, pNomads->pDocRoot);
	
	apr_pool_userdata_get(&(pNomads->pConf), szLine, SysPool);
	if (pNomads->pConf != NULL) return TRUE;
	
	pNomads->pConf = (struct CONFIG *)apr_pcalloc(SysPool, sizeof(struct CONFIG));
	apr_pool_userdata_set(pNomads->pConf, szLine, NULL, SysPool);

	pNomads->pConf->tblLinks = apr_table_make(SysPool, 100);

	strcpy(szPath, pNomads->pDocRoot);
	strcat(szPath, "/inomads.conf");
	
	pNomads->pConf->pLaunchPswd = "";
	pNomads->pConf->nLaunchPort = 4093;
	pNomads->pConf->nTimeout = 60 * 5;			/* Five minute default */
	pNomads->pConf->pRmtRoot = "";
	
	if (apr_file_open(&pfConf, szPath, APR_READ, APR_OS_DEFAULT, TmpPool) == APR_SUCCESS)
	{	for(;;)
		{	if (apr_file_gets(szLine, sizeof(szLine), pfConf) == APR_EOF) break;
			pDlm = strchr(szLine, '\r');
			if (pDlm) *pDlm = 0;
			
			pDlm = strchr(szLine, '\n');
			if (pDlm) *pDlm = 0;
			
			if ((memcmp(szLine, "DIRECTORY=", 10) == 0) && (szLine[10] != 0))
				pNomads->pConf->pDir = apr_pstrdup(SysPool, szLine + 10);
			else
			if ((memcmp(szLine, "PXPLUS_EXEC=", 12) == 0) && (szLine[12] != 0))
				pNomads->pConf->pExe = apr_pstrdup(SysPool, szLine + 12);
			else
			if ((memcmp(szLine, "PXPLUS_LIB=", 11) == 0) && (szLine[11] != 0))
				pNomads->pConf->pLib = apr_pstrdup(SysPool, szLine + 11);
			else
			if ((memcmp(szLine, "SYSHTML_DIR=", 12) == 0) && (szLine[12] != 0))
				pNomads->pConf->pHtml = apr_pstrdup(SysPool, szLine + 12);
			else
			if ((memcmp(szLine, "LAUNCH_IP=", 10) == 0) && (szLine[10] != 0))
				pNomads->pConf->pLaunchIP = apr_pstrdup(SysPool, szLine + 10);
			else
			if ((memcmp(szLine, "LAUNCH_PORT=", 12) == 0) && (szLine[12] != 0))
				pNomads->pConf->nLaunchPort = atoi(szLine + 12);
			else
			if ((memcmp(szLine, "LAUNCH_PSWD=", 12) == 0) && (szLine[12] != 0))
				pNomads->pConf->pLaunchPswd = apr_pstrdup(SysPool, szLine + 12);
			else
			if ((memcmp(szLine, "LAUNCH_ROOT=", 12) == 0) && (szLine[12] != 0))
				pNomads->pConf->pRmtRoot = apr_pstrdup(SysPool, szLine + 12);
			else
			if ((memcmp(szLine, "TCP_TIMEOUT=", 12) == 0) && (szLine[12] != 0))
				pNomads->pConf->nTimeout = atoi(szLine + 12);
		}

		apr_file_close(pfConf);
	}
		
	if (pNomads->pConf->pDir == NULL)
		pNomads->pConf->pDir = apr_pstrdup(SysPool, pNomads->pDocRoot);
	
	if (pNomads->pConf->pExe == NULL) 
	{	strcpy(szLine, pNomads->pConf->pDir);
		pTemp = szLine;
		for(;;)
		{	pDlm = strchr(pTemp, '/');
			if (pDlm == NULL) pDlm = strchr(pTemp, '\\');
			if (pDlm == NULL) break;
			
			pTemp = pDlm + 1;			
			if (memcmp(pTemp, "lib", 3) == 0)
			{	strcpy(pTemp, PXPLUS_EXE);
				pNomads->pConf->pExe = apr_pstrdup(SysPool, szLine);
				break;
			}
		}
	}
	
	if (pNomads->pConf->pLib == NULL)
	{	strcpy(szPath, pNomads->pConf->pExe);
		
		pDlm = strrchr(szPath, '/');
		if (pDlm == NULL) pDlm = strrchr(szPath, '\\');
		
		if (pDlm) *pDlm = 0;
		pNomads->pConf->pLib = apr_psprintf(SysPool, "%s/lib", szPath);
	}
		
	if (pNomads->pConf->pHtml == NULL)
		pNomads->pConf->pHtml = apr_psprintf(SysPool, "%s/syshtml", pNomads->pDocRoot);

	return TRUE;
}

/* Parse_Args(pNomads)
 *
 *		Parse the reequest argument list into the tblArgs array
 */
static int Parse_Args(struct INOMADS *pNomads)
{	char szArgs[HUGE_STRING_LEN];
	
	if (Request->args)
		strcpy(szArgs, Request->args);
	else
		strcpy(szArgs, "");
	
	pNomads->tblArgs = TblParse(pNomads, szArgs);
	return OK;
}

/* Read_Contents(pNomads)
 *
 *		If the submission is a form or file upload read the contents of the submission
 *		into the pContents (len in lContents) for subsequent processing.
 *		If no data then set contents to "" and length to zero.
 */
static int Read_Contents(struct INOMADS *pNomads)
{	char				bfrTemp[HUGE_STRING_LEN];
    int					nRead;
    int					nOfs;
    
    pNomads->pContent = "";
    pNomads->lContent = 0;
    
	if (Request->method_number != M_POST) return TRUE;
	
	pNomads->pType = (char *)apr_table_get(Request->headers_in, "Content-Type");

	if (apr_strnatcasecmp(pNomads->pType, "application/x-www-form-urlencoded") != 0) 
	{	if (strlen(pNomads->pType) <= 19) return FALSE;
		if (memicmp(pNomads->pType, "multipart/form-data", 19) != 0) return TRUE;
		pNomads->bUpload = TRUE;
	}

	if ((RetCode = ap_setup_client_block(Request, REQUEST_CHUNKED_ERROR)) != OK)
		return Send_Error(pNomads, "Cannot access POSTed data");
	
	if (!ap_should_client_block(Request)) return TRUE;

	if (Request->remaining > 10*1024*1024) 
		return Send_Error(pNomads, "Upload failure: File size exceeds allowable"); 

	pNomads->lContent = (long)Request->remaining;
	pNomads->pContent = apr_pcalloc(TmpPool, pNomads->lContent + 1);
	
	nOfs = 0;
		
	while ((nRead = ap_get_client_block(Request, bfrTemp, sizeof(bfrTemp))) > 0)
	{	if ((nOfs + nRead) > pNomads->lContent)
		{	nRead = pNomads->lContent - nOfs;
		}

		if (nRead) memcpy(pNomads->pContent + nOfs, bfrTemp, nRead);
		nOfs += nRead;
	}

	return TRUE;
}

/* Parse_Contents(pNomads)
 *
 *		If the content was a file upload then parse the uploaded data otherwise simply
 *		parse the contents to construct the table of form fields.  
 *
 *		Note that the contents may be "" however we still parse the data to that the 
 *		form array gets created
 */
static int Parse_Contents(struct INOMADS *pNomads)
{	if (pNomads->bUpload) return Parse_Upload(pNomads);

	pNomads->tblForm = TblParse(pNomads, pNomads->pContent);
	return TRUE;
}

/* Parse_Upload(pNomads)
 *
 *		This function is called if the request was a file upload.
 *		It will parse the received contents in order to extract the file and write
 *		it to the session directory.
 */
static int Parse_Upload(struct INOMADS *pNomads)
{	char			*pTemp;
	char			*pToken;
    char			*pFile;
    char			*pStart;
    char			*pName;
    char			*pBound;
	char			*pLine;
    apr_size_t		lBound = 0;
	char			szPath[MAX_PATH_NAME];

	char			*pData;
	apr_size_t		lData;
	apr_file_t		*pfFile;
	
    pFile = (char *)apr_table_get(pNomads->tblArgs,"_upload");
    if (pFile == NULL) return Send_Error(pNomads, "Upload failure: No upload name"); 
    
    if (strlen(pFile) == 0) return TRUE;
    
   	pNomads->pSession = (char *)apr_table_get(pNomads->tblArgs,"_session");
    if (pNomads->pSession == NULL) return Send_Error(pNomads, "Upload failure: No session identifed"); 

    pTemp = pNomads->pType;
    
    for (;pTemp != NULL;)
    {	pTemp = strchr(pTemp,'b');
		if (pTemp == NULL) break;
		if (memicmp(pTemp, "boundary=", 9) == 0) break;
		pTemp++;
		break;
	}
	
	if (!pTemp) return Send_Error(pNomads, "Upload failure: Invalid boundary in Content-type"); 

	pBound = pTemp + 9;
	pTemp = strchr(pBound, ';');
	if (pTemp)
		lBound = (int)(pTemp - pBound);
	else
		lBound = (int)strlen(pBound);

	/* Split input based on Boundaries
	** Set pointers/length to Content sent with the request
	*/
	
	lData = pNomads->lContent;
	pData = pNomads->pContent;
	
	pStart = NULL;			/* File not found */
	pName = NULL;			/* File name not found */

	for(;;)
	{	/* Find next Boundary */
	
		for(;lData > 0;)
		{	pTemp = memchr(pData, '-', lData);
			if (pTemp == NULL) 
			{	lData = 0;
				break;
			}
			
			lData -= (int)(pTemp - pData);
			pData = pTemp;
					
			if (lData < lBound+2)
			{	lData = 0;
				break;
			}
			
			if ((pData[1] == '-') && (memcmp(pData+2, pBound, lBound) == 0)) break;

			pData++;
			lData--;
		}
		
		if (!lData) break;
		if (pStart) break;			/* New boundary after start */
		
		/* Break out the records */
		for(;lData > 0;)
		{	pLine = pData;
		
			pTemp = memchr(pData, '\n', lData);
			if (pTemp == NULL) 
			{	lData = 0;
				break;
			}

			lData -= (int)(pTemp - pData) + 1;
			pData = pTemp + 1;

			*pTemp = 0;
			pTemp--;
			if (*pTemp == '\r') *pTemp = 0;

			if (pLine[0] == 0)
			{	pStart = pData;
				break;
			}

			pTemp = strchr(pLine, ':');
			if (pTemp != NULL)
			{	*(pTemp++) = 0;
				if (apr_strnatcasecmp("Content-Disposition", pLine) == 0)
				{	strcat(pTemp, ";");
				
					for (pTemp = apr_strtok(pTemp, ";", &pToken); pTemp != NULL; pTemp = apr_strtok(NULL, ";", &pToken))
					{	size_t				nLen;
					
						/* Strip leading and trailing spaces from token */
					
						for (;*pTemp == ' ';pTemp++);
						for (nLen = strlen(pTemp);nLen > 0;)
						{	if (pTemp[--nLen] != ' ') break;
							pTemp[nLen] = 0;
						}
					
						if (memicmp("filename=", pTemp, 9) == 0)
						{	pName = pTemp + 9;
							if (pName[0] == '"')
							{	pName++;
								pTemp = strchr(pName, '"');
								if (pTemp) *pTemp = 0;
							}
							UnEscapeURL(pName);
							break;
						}
					}
				}
				continue;
			}
		}
	}	

	if (!pName) return Send_Error(pNomads, "Upload failure: Cannot locate file name"); 
	if (!pStart) return Send_Error(pNomads, "Upload failure: No data for file"); 
	
	pNomads->tblForm = apr_table_make(TmpPool, 10);
	apr_table_merge(pNomads->tblForm, "upload..filename", pName);							

	if (*(pData-1) == '\n')
	{	pData--;			/* Toss the trailing Line feed */
		if (*(pData-1) == '\r') pData--;
	}

	lData = (int)(pData - pStart);

	strcpy(szPath, pNomads->pDocRoot);
	strcat(szPath, "/tmp/");
	strcat(szPath, pNomads->pSession);
    strcat(szPath, "/");
    strcat(szPath, pFile);

	if (apr_file_open(&pfFile, szPath, APR_WRITE | APR_CREATE | APR_TRUNCATE | APR_BINARY, APR_OS_DEFAULT, TmpPool) != APR_SUCCESS)
		return Send_Error(pNomads, apr_psprintf(TmpPool, "Upload failure: Cannot open file on server.<br>(%s)", szPath)); 

	apr_file_write(pfFile, pStart, &lData);
	apr_file_close(pfFile);

	return TRUE;
}

static int Load_session(struct INOMADS *pNomads)
{	int				nIdx;
	int			bNew;
	char			szPath[MAX_PATH_NAME];
	char			*pTmp;
	apr_file_t		*pfFile;
	struct TBLINFO	tblInfo;
	
	/* See if we have a session ID */

	pNomads->pSession = (char *)apr_table_get(pNomads->tblArgs,"_session");
	if ((pNomads->pSession == NULL) &&  (pNomads->tblForm))
		pNomads->pSession = (char *)apr_table_get(pNomads->tblForm, "_session");

	if (pNomads->pSession != NULL)
	{	bNew = FALSE;
	}
	else
	{	bNew =  TRUE;
		pNomads->pSession = apr_pcalloc(TmpPool, 13);
		apr_generate_random_bytes((unsigned char *)pNomads->pSession, 12);

		for (nIdx=0; nIdx<12; nIdx++)
		{	pNomads->pSession[nIdx] = 97 + (((unsigned char *)pNomads->pSession)[nIdx] % 26);
		}
	}
	
	strcpy(szPath, pNomads->pDocRoot);
	strcat(szPath, "/tmp/");
	strcat(szPath, pNomads->pSession);
	pNomads->pSesDiry = apr_pstrdup(TmpPool, szPath);
	
	if (!bNew) return TRUE;
	
	if (!Load_Inid(pNomads)) return FALSE;

	/* Create the directory */
	
	if (apr_dir_make(szPath, APR_OS_DEFAULT, TmpPool) != APR_SUCCESS)
	{	pNomads->pMsg = apr_psprintf(TmpPool, "Unable to create working directory: %s", szPath);
		Send_SysHtml(pNomads, "sys_error.htm");
		return FALSE;
	}
	
	strcat(szPath, "/.info");

	pfFile = NULL;
	if (apr_file_open(&pfFile, szPath, APR_WRITE | APR_CREATE | APR_TRUNCATE, APR_OS_DEFAULT, TmpPool) != APR_SUCCESS)
	{	pNomads->pMsg = apr_psprintf(TmpPool, "Unable to open: %s", szPath);
		Send_SysHtml(pNomads, "sys_error.htm");
		return FALSE;
	}
	
	apr_file_printf(pfFile,"HTTP_USER_AGENT=%s" NL,	apr_table_get(Request->headers_in,"User-Agent"));

	apr_file_printf(pfFile,"HTTP_HOST=%s" NL,		apr_table_get(Request->headers_in,"Host"));
	apr_file_printf(pfFile,"REMOTE_ADDR=%s" NL,		Request->connection->remote_ip);
	apr_file_printf(pfFile,"SERVER_ADDR=%s" NL,		Request->connection->local_ip);
	apr_file_printf(pfFile,"SERVER_PORT=%d" NL,		ap_get_server_port(Request));

	if (apr_table_get(Request->subprocess_env, "HTTPS"))
		nIdx = 1;
	else
		nIdx = 0;
			
	apr_file_printf(pfFile,"SERVER_SECURE=%d" NL,	nIdx);
	
	apr_file_printf(pfFile,"DOCUMENT_ROOT=%s" NL,	pNomads->pRmtRoot);

	pTmp = strchr(Request->the_request, '/') + 1;

	for (nIdx = 0;;)
	{	if ((*pTmp == ' ') || (*pTmp == '?') || (*pTmp == 0)) break;
		szPath[nIdx++] = *(pTmp++);
	}
	
	if (szPath[nIdx-1] == '/') nIdx--;
	szPath[nIdx] = 0;
	if (nIdx == 0) strcpy(szPath, "inomads");
	
	apr_file_printf(pfFile,"WEB_URL=%s" NL,			szPath);
	apr_file_printf(pfFile,"INID=%s" NL NL,			pNomads->pInid);

	memset(&tblInfo, 0, sizeof(tblInfo));
		
	apr_table_do(CountArg, &tblInfo, pNomads->tblArgs, NULL);
	apr_file_printf(pfFile,"<Args count=%d>" NL,		tblInfo.nKeys);
	
	apr_table_do(PrintArg, pfFile, pNomads->tblArgs, NULL);
	apr_file_printf(pfFile,"</Args>" NL);

	apr_file_close(pfFile);
	
	if (pNomads->pConf->pLaunchIP)
	{	if (!Spawn_ViaTcp(pNomads)) return FALSE;
	}
	else
	{	if (!Spawn_Direct(pNomads)) return FALSE;
	}
   
    strcpy(szPath, pNomads->pSesDiry);
    strcat(szPath, "/.link");
    
	for (nIdx = 200; nIdx > 0; nIdx--)
	{	apr_sleep(apr_time_from_sec(5) / 100);
		if (access(szPath, 0) == 0) break;
	}
	
	if (nIdx <= 0)
	{	return Send_Error(pNomads, "Unable to connect to newly spawned task");
	}

	return TRUE;
}

/* Spawn_Direct(pNomads)
 *
 *		This function will directly spawn the process on the save server/service
 */
static int Spawn_Direct(struct INOMADS *pNomads)
{	const char			*szArgs[10];

	apr_procattr_t	*procAttr;
	apr_proc_t		procInfo;
	apr_exit_why_e	procWhy;

	/* Spawn the process */
	
	if (apr_procattr_create(&procAttr, TmpPool) != APR_SUCCESS)
	{	return Send_Error(pNomads, "Unable to create process attribute structure");
	}
	
	if (pNomads->pConf->pDir)
	{	if (apr_procattr_dir_set(procAttr, pNomads->pConf->pDir) != APR_SUCCESS)
		{	return Send_Error(pNomads, apr_psprintf(TmpPool, "Unable to change directory to: %s", pNomads->pConf->pDir));
		}
	}
	
	apr_procattr_detach_set(procAttr, TRUE);
	apr_procattr_cmdtype_set(procAttr, APR_PROGRAM_ENV);
	
	szArgs[0] = pNomads->pConf->pExe;			/* PxPlus EXE */
	szArgs[1] = "-mn";
#if defined(_WINDOWS	)
	szArgs[2] = "'*plus/inomads/inomads'";
	szArgs[3] = "-arg";
	szArgs[4] = apr_psprintf(TmpPool, "'%s;%s'", pNomads->pSession, pNomads->pRmtRoot);
#else
	szArgs[2] = "*plus/inomads/inomads";	
	szArgs[3] = "-arg";
	szArgs[4] = apr_psprintf(TmpPool, "%s;%s", pNomads->pSession, pNomads->pRmtRoot);
#endif
	szArgs[5] = NULL;

	memset(&procInfo, 0, sizeof(procInfo));
	
	if (apr_proc_create(&procInfo, pNomads->pConf->pExe, szArgs, NULL, procAttr, TmpPool) != APR_SUCCESS)
		return Send_Error(pNomads, "Unable to create/launch process");
		
	apr_proc_wait(&procInfo, NULL, &procWhy, APR_NOWAIT);
	
	return TRUE;
}


/* Spawn_ViaTcp(pNomads)
 *
 *		This function will call the specified launch port and request the task be spawned
 */
static int Spawn_ViaTcp(struct INOMADS *pNomads)
{	char			szSpawn[512];
	apr_size_t		lTemp;
	apr_size_t		lSpawn;
	
	apr_sockaddr_t	*pSockAddr;
	apr_socket_t	*pSocket;

	memset(szSpawn, 0, sizeof(szSpawn));
	memcpy(szSpawn, "\xDE\xAD\xBE\xEF", 4);
	lSpawn = 4;
	
	lTemp = strlen(pNomads->pConf->pLaunchPswd) + 1;
	memcpy(szSpawn+lSpawn, pNomads->pConf->pLaunchPswd, lTemp);
	lSpawn += lTemp;
	
	lTemp = strlen(pNomads->pSession) + 1;
	memcpy(szSpawn+lSpawn, pNomads->pSession, lTemp);
	lSpawn += lTemp;
	
	lTemp = strlen(pNomads->pRmtRoot) + 1;
	memcpy(szSpawn+lSpawn, pNomads->pRmtRoot, lTemp);
	lSpawn += lTemp;

	if ((apr_sockaddr_info_get(&pSockAddr, pNomads->pConf->pLaunchIP, APR_INET, pNomads->pConf->nLaunchPort, 0, TmpPool) ||
		 apr_socket_create(&pSocket, pSockAddr->family, SOCK_STREAM, APR_PROTO_TCP, TmpPool) ||
		 apr_socket_opt_set(pSocket, APR_SO_NONBLOCK, 1) ||
		 apr_socket_timeout_set(pSocket, apr_time_from_sec(5)) ||
		 apr_socket_connect(pSocket, pSockAddr) ||
		 apr_socket_opt_set(pSocket, APR_SO_NONBLOCK, 0) ||
		 apr_socket_timeout_set(pSocket, apr_time_from_sec(5)) ||
		 apr_socket_send(pSocket, szSpawn, &lSpawn)) != APR_SUCCESS)
		{	return Send_Error(pNomads, "Unable to connect process launch task");
		}
		
	lSpawn = 2;
	if ((apr_socket_recv(pSocket, szSpawn, &lSpawn) != APR_SUCCESS) ||
		(lSpawn != 2) ||
		(memcmp(szSpawn, "OK", 2) != 0))
		{	apr_socket_close(pSocket);
			return Send_Error(pNomads, "Invalid launcher response");
		}
	
	apr_socket_close(pSocket);
	return TRUE;
}	

static int CountArg(void* pInfo, const char* pKey, const char* pValue)
{	struct TBLINFO	*pTbl;
	int				nLen;
	
	pTbl = (struct TBLINFO *)pInfo;

	pTbl->nKeys++;

	nLen = (int)strlen(pKey);		
	pTbl->lenKeys += nLen + 2;

	if (pTbl->pKeys)
	{	*(pTbl->pKeys++) = ',';
		
		memcpy(pTbl->pKeys, pKey, nLen);
		pTbl->pKeys += nLen;
		
		*(pTbl->pKeys++) = '$';
	}

	nLen = (int)strlen(pValue);
	pTbl->lenData += nLen + 1;
	if (pTbl->pData)
	{	*(pTbl->pData++) = 0;
	
		memcpy(pTbl->pData, pValue, nLen);
		pTbl->pData += nLen;
	}

	return 1;	
}

static int PrintArg(void* pFile, const char* pKey, const char* pValue)
{	apr_file_printf((apr_file_t *)pFile, "%s=%s" NL, pKey, pValue);
	return 1;	
}

static void Exchg_data(struct INOMADS *pNomads)
{	char			szPath[MAX_PATH_NAME];
	char			szLine[80];
	char			*pWrk;
	char			*pExit;
	char			*pSeq;
	char			*pRemoteIP;
	char			*pSend;
	char			*pData;
	apr_size_t		lenData;

	char			bfrRecv[8192];
	apr_size_t		lenRecv;
	int				bMore;
	int				bExit;
	
	int				nPort;
	struct TBLINFO	tblInfo;
	
	apr_file_t		*pfFile;
	apr_sockaddr_t	*pSockAddr;
	apr_socket_t	*pSocket;
	apr_interval_time_t	nTimeout;
	
	pNomads->pLink = (char *)apr_table_get(pNomads->pConf->tblLinks, pNomads->pSession);
	if (pNomads->pLink)
	{	strcpy(szLine, pNomads->pLink);
	}
	else
	{	strcpy(szPath, pNomads->pSesDiry);
		strcat(szPath, "/.link");

		if (apr_file_open(&pfFile, szPath, APR_READ, APR_OS_DEFAULT, TmpPool) != APR_SUCCESS)
		{	Send_SysHtml(pNomads, "sys_no_ses.htm");
			return;
		}	

		if (apr_file_gets(szLine, sizeof(szLine), pfFile) != APR_SUCCESS)
		{	Send_SysHtml(pNomads, "sys_no_ses.htm");
			return;
		}	
		
		apr_file_close(pfFile);
	
		pWrk = strchr(szLine, '\r');
		if (pWrk) *pWrk = 0;
		pWrk = strchr(szLine, '\n');
		if (pWrk) *pWrk = 0;

		apr_table_set(pNomads->pConf->tblLinks, pNomads->pSession, szLine);
		pNomads->pLink = (char *)apr_table_get(pNomads->pConf->tblLinks, pNomads->pSession);
	}

	pWrk = strchr(szLine, ';');
	if (!pWrk)
	{	pNomads->pMsg = apr_psprintf(TmpPool, "Invalid Connection string: '%s'", szLine);
		Send_SysHtml(pNomads, "sys_error.htm");
		return;
	}

	*pWrk = 0;
	nPort = atoi(pWrk+1);
	
	if (pNomads->pConf->pLaunchIP) strcpy(szLine, pNomads->pConf->pLaunchIP);

	/* Generate data to pass to system */
	
	pRemoteIP = Request->connection->remote_ip;
	if (pRemoteIP == NULL) pRemoteIP = "0.0.0.0";

	pSeq = (char *)apr_table_get(pNomads->tblArgs,"_seq");
	
	if ((pSeq == NULL) && (pNomads->tblForm))
		pSeq = (char *)apr_table_get(pNomads->tblForm, "_seq");
	
	if (pSeq == NULL)
		pSeq = "0000";
	else
	if (strlen(pSeq) != 4)
	{	pNomads->pMsg = apr_pstrcat(TmpPool, "Invalid sequence number:", pSeq, NULL);
		Send_SysHtml(pNomads, "sys_error.htm");
		return;
	}

	if (strcmp(pSeq, "Rscu") == 0)
	{	if (!pNomads->pInid) Load_Inid(pNomads);
		apr_table_merge(pNomads->tblArgs, "_inid", pNomads->pInid);
	}
	
	memset(&tblInfo, 0, sizeof(tblInfo));
	apr_table_do(CountArg, &tblInfo, pNomads->tblArgs, NULL);
	apr_table_do(CountArg, &tblInfo, pNomads->tblForm, NULL);

	/* Allocate memory required for message */	

	lenData = 8 + 1 + (int)strlen(pRemoteIP) + 1 + tblInfo.lenKeys + tblInfo.lenData + 1; 
	pSend = pData = apr_pcalloc(TmpPool, lenData);	

	/* Insert length & sequence number for 8+1 bytes */

	sprintf(pData, "%04d%s", lenData - 8, pSeq);
	
	pData += 8 + 1;
	
	/* Insert IP address */
	
	strcpy(pData, pRemoteIP);
	pData += (int)strlen(pRemoteIP) + 1;
	
	/* Now insert keys and data values */

	tblInfo.pKeys = pData;
	tblInfo.pData = pData + tblInfo.lenKeys;
	
	tblInfo.lenData = tblInfo.lenKeys = tblInfo.nKeys = 0;
	
	apr_table_do(CountArg, &tblInfo, pNomads->tblArgs, NULL);
	apr_table_do(CountArg, &tblInfo, pNomads->tblForm, NULL);

	/* Okay now lets open the socket */
	
	if (pNomads->pConf->nTimeout == 0)
		nTimeout = -1;
	else
		nTimeout = apr_time_from_sec(pNomads->pConf->nTimeout);
		
	if ((apr_sockaddr_info_get(&pSockAddr, szLine, APR_INET, nPort, 0, TmpPool) ||
		 apr_socket_create(&pSocket, pSockAddr->family, SOCK_STREAM, APR_PROTO_TCP, TmpPool) ||
		 apr_socket_opt_set(pSocket, APR_SO_NONBLOCK, 1) ||
		 apr_socket_timeout_set(pSocket, apr_time_from_sec(60)) ||
		 apr_socket_connect(pSocket, pSockAddr) ||
		 apr_socket_opt_set(pSocket, APR_SO_NONBLOCK, 0) ||
		 apr_socket_timeout_set(pSocket, nTimeout) ||
		 apr_socket_send(pSocket, pSend, &lenData)) != APR_SUCCESS)
		{	Send_SysHtml(pNomads, "sys_no_ses.htm");
			return;
		}	
	
	bExit = TRUE;
	memset(szPath, 0, sizeof(szPath));
		
	for(;;)
	{	lenRecv = 2;
	
		if ((apr_socket_recv(pSocket, bfrRecv, &lenRecv) != APR_SUCCESS) || (lenRecv != 2))
		{	pNomads->pMsg = "Reception error on header (Timeout or Disconnect)";
			Send_SysHtml(pNomads, "sys_error.htm");
			return;
		}	
	
		if (bfrRecv[0] & 0x80)
			bMore = TRUE;
		else
			bMore = FALSE;
			
		lenData = ((bfrRecv[0] & 0x7F) << 8) | (bfrRecv[1] & 0x00FF);
		
		if ((lenData < 0) || (lenData > 10240))
		{	lenRecv = 40;
			apr_socket_recv(pSocket, bfrRecv, &lenRecv);
			
			pNomads->pMsg = apr_psprintf(TmpPool, "Header length indicated %d bytes -- {%40.40s}", lenData, bfrRecv);
			Send_SysHtml(pNomads, "sys_error.htm");
			bExit = FALSE;
			break;
		}
					
		for (;lenData > 0;)
		{	apr_status_t		stsRecv;
		
			lenRecv = lenData;
			if (lenRecv > sizeof(bfrRecv)) lenRecv = sizeof(bfrRecv);
			
			stsRecv = apr_socket_recv(pSocket, bfrRecv, &lenRecv);

			if (stsRecv != APR_SUCCESS) 
			{	pNomads->pMsg = apr_psprintf(TmpPool, "Unable to receive full block - Want %u got %u - sts=%d", 
					lenData, lenRecv, stsRecv);
				Send_SysHtml(pNomads, "sys_error.htm");
				return;
			}	
			
			if (lenRecv)
			{	if ((bExit) && (lenRecv >= 5) && (memcmp(bfrRecv, "exit:", 5) == 0))
				{	memcpy(szPath, bfrRecv + 5, lenRecv - 5);
				}
				else
				{	bExit = FALSE;
					ap_rwrite(bfrRecv, (int)lenRecv, Request);
				}
				
				lenData -= lenRecv;
			}
		}
	
		if (!bMore) break;
	}
	
	apr_socket_close(pSocket);
	if (bExit != TRUE) return;
	
	pExit = strchr(szPath, 0x09);
	if (pExit == NULL)
		pExit = "";
	else
	{	*pExit = 0;	
		pExit++;
	}
		
	pWrk = strchr(pExit, 0x09);
	if (pWrk) *pWrk = 0;
	
	if (szPath[0] != 0) pNomads->pAjax = szPath;
	
	if (strcmp(pExit, "blank") == 0)
	{	Send_SysHtml(pNomads, "sys_blank.htm");
		return;
	}
	
	if (strcmp(pExit, "close") == 0)
	{	Send_SysHtml(pNomads, "sys_close.htm");
		return;
	}

	if (pExit[0] != 0)
	{	pNomads->pMsg = pExit;
	
		if (!pNomads->pAjax) pNomads->pAjax = "";
		pNomads->pAjax = apr_psprintf(TmpPool, "%s\nsUrl = '%s';", pNomads->pAjax, pExit);
		Send_SysHtml(pNomads, "sys_xfer.htm");
		return;
	}
	
	Send_SysHtml(pNomads,"sys_endapp.htm");
	return;
}

static int Send_Error(struct INOMADS *pNomads, char *pMsg)
{	pNomads->pMsg = pMsg;
	Send_SysHtml(pNomads, "sys_error.htm");
	return FALSE;
}
	
static void Send_SysHtml(struct INOMADS *pNomads, char *sHtmlFile)
{	char			szPath[MAX_PATH_NAME];
	char			szLine[1000];
	apr_file_t		*pfHtml;
	
	if (pNomads->pLink)
	{	apr_table_unset(pNomads->pConf->tblLinks, pNomads->pSession);
		pNomads->pLink = NULL;
	}
	
	sprintf(szPath, "%s/%s", pNomads->pConf->pHtml, sHtmlFile);

	if (apr_file_open (&pfHtml, szPath, APR_READ, APR_OS_DEFAULT, TmpPool) != APR_SUCCESS)
	{	ap_rprintf(Request, "<html><body>Cannot find %s</body></html>", szPath);
		return;
	}
	
	for(;;)
	{	if (apr_file_gets(szLine, sizeof(szLine), pfHtml) == APR_EOF) break;
		if (memcmp(szLine, "//Ajax", 6) == 0)
		{	if (pNomads->pAjax) ap_rputs(pNomads->pAjax, Request);
			continue;
		}
		if (memcmp(szLine, "//Meta", 6) == 0)
		{	if (pNomads->pMeta) ap_rputs(pNomads->pMeta, Request);
			continue;
		}
		if (memcmp(szLine, "//Msg", 5) == 0)
		{	if (pNomads->pMsg) ap_rputs(pNomads->pMsg, Request);
			continue;
		}
		
		ap_rputs(szLine, Request);
	}
	apr_file_close(pfHtml);
	return;
}

static int Load_Inid(struct INOMADS *pNomads)
{	char			*pCookies;
	char			bfrTemp[500];
	char			*pTemp;
	char			*pToken;
	char			*pDlm;
	
	char			*pInidCookie;
	char			szLastInid[100];
	char			szPath[MAX_PATH_NAME];

	long			nLastInid;
	long			nInid;
		
	apr_file_t		*pfInid;
	apr_time_exp_t	tmLocal;
 
	pCookies = (char *)apr_table_get(Request->headers_in, "Cookie");
	if (pCookies != NULL)
	{	if ((strlen(pCookies)+1) > sizeof(bfrTemp))
			return Send_Error(pNomads, "Too many cookies");
			
		strcpy(bfrTemp, pCookies);
		pTemp = bfrTemp;
	
		for (pTemp = apr_strtok(pTemp, " ;", &pToken); pTemp != NULL; pTemp = apr_strtok(NULL, " ;", &pToken))
		{	pDlm = strchr(pTemp, '=');
			if (pDlm)
			{	*(pDlm++) = 0;
				if (apr_strnatcasecmp(pTemp, "inid") == 0) 
				{	UnEscapeURL(pDlm);
					pNomads->pInid = apr_pstrdup(TmpPool, pDlm);
					break;
				}
			}
		}
	}
	
	if (pNomads->pInid != NULL) return TRUE;

	strcpy(szPath, pNomads->pDocRoot);
	strcat(szPath, "/lastinid.conf");
	nLastInid = 0;
	
	pfInid = NULL;
	
	if (apr_file_open (&pfInid, szPath, APR_READ, APR_OS_DEFAULT, TmpPool) == APR_SUCCESS)
	{	apr_file_gets(szLastInid, sizeof(szLastInid), pfInid);
		szLastInid[10] = 0;
		nLastInid = atoi(szLastInid);
		apr_file_close(pfInid);
	}
	
	apr_time_exp_lt(&tmLocal, apr_time_now());
	
	nInid = tmLocal.tm_year % 100;
	nInid = (nInid * 100) + tmLocal.tm_mon + 1;
	nInid = (nInid * 100) + tmLocal.tm_mday;
	nInid = (nInid * 100) + tmLocal.tm_hour;
	nInid = (nInid * 100) + tmLocal.tm_min;
			
	if (nInid <= nLastInid) nInid = nLastInid + 1;
	
	pNomads->pInid = apr_pcalloc(TmpPool, 13);
	sprintf(pNomads->pInid, "%010ld", nInid);
	
	if (apr_file_open (&pfInid, szPath, APR_WRITE | APR_CREATE | APR_TRUNCATE, APR_OS_DEFAULT, TmpPool) != APR_SUCCESS)
	{	pNomads->pMsg = apr_psprintf(TmpPool, "Unable update file %s with browser ID", szPath);
		Send_SysHtml(pNomads, "sys_error.htm");
		return FALSE;
	}

	apr_file_printf(pfInid, "%s" NL, pNomads->pInid);
	apr_file_close(pfInid);

	tmLocal.tm_year += 5;
	
	pInidCookie = apr_psprintf(TmpPool,	"inid=%s; path=/; expires=%s, %.2d-%s-%.2d %.2d:%.2d:%.2d GMT",
						pNomads->pInid,
						apr_day_snames[tmLocal.tm_wday],
						tmLocal.tm_mday,
						apr_month_snames[tmLocal.tm_mon],
						tmLocal.tm_year % 100,
						tmLocal.tm_hour, 
						tmLocal.tm_min,
						tmLocal.tm_sec);
	
	apr_table_setn(Request->headers_out, "Set-Cookie", pInidCookie);
	return TRUE;
}

static apr_table_t *TblParse(struct INOMADS *pNomads, const char *data)
{	const char *key, *val;
	apr_table_t *tab;
	
	tab = apr_table_make(TmpPool, 20);

    while(*data && (val = ap_getword(TmpPool, &data, '&')))
    {	key = ap_getword(TmpPool, &val, '=');
		UnEscapeURL((char*)key);
		UnEscapeURL((char*)val);
		apr_table_merge(tab, key, val);
    }

    return tab;
}

static int UnEscapeURL(char *pVal)
{	int			oInp;
	int			oOut;
	char		cChar, cChar2;

	oInp = oOut = 0;
	for (;;)
	{	cChar = pVal[oInp++];
		if (cChar == '+') cChar = ' ';
		if (cChar != '%')
		{	pVal[oOut++] = cChar;
			if (cChar == 0) return TRUE;
			continue;
		}
			
		cChar = pVal[oInp++];
		if ((cChar >= '0') && (cChar <= '9'))
			cChar2 = cChar - '0';
		else
		if ((cChar >= 'A') && (cChar <= 'F'))
			cChar2 = cChar - 'A' + 10;
		else
		if ((cChar >= 'a') && (cChar <= 'f'))
			cChar2 = cChar - 'a' + 10;
		else
			break;
			
		cChar = pVal[oInp++];
		if ((cChar >= '0') && (cChar <= '9'))
			cChar = cChar - '0';
		else
		if ((cChar >= 'A') && (cChar <= 'F'))
			cChar = cChar - 'A' + 10;
		else
		if ((cChar >= 'a') && (cChar <= 'f'))
			cChar = cChar - 'a' + 10;
		else
			break;
		
		pVal[oOut++] = cChar2 << 4 | cChar;
	}
	
	return FALSE;
}			

/* Hook our handler into Apache at startup */

static void inomads_hooks(apr_pool_t* pool)
{	ap_hook_handler(inomads_handler, NULL, NULL, APR_HOOK_MIDDLE) ;
}

module AP_MODULE_DECLARE_DATA inomads_module = {
        STANDARD20_MODULE_STUFF,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        inomads_hooks
};
